# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Inverse geocoding module
------------------------
"""

import copy
import numpy as np
import arepytools.constants as cst
from arepytools import _utils

_VELOCITY_SCENE = np.asarray([0, 0, 0])

_NEWTON_TOLERANCE = 1.e-8
_MAX_ITERATIONS = 8
_ZERO_TOLERANCE = 1.e-8


def inverse_geocoding_monostatic(gso, point: np.ndarray, frequency_doppler_centroid: float,
                                 wavelength: float) -> (list, list):
    """Perform monostatic inverse geocoding

    :param gso: general sar orbit
    :param point: a point (xyz) coordinate
    :param frequency_doppler_centroid: doppler centroid
    :param wavelength:  wavelength

    More than one solution is possible: that's why lists are returned.

    :return: azimuth_times, range_times. Two lists with the solutions.
    """
    _check_inverse_geocoding_input(point, frequency_doppler_centroid, wavelength)

    interval_ids = _inverse_geocoding_monostatic_init(gso, point.reshape((3, 1)), frequency_doppler_centroid,
                                                      wavelength)

    azimuth_times, range_times = list(), list()
    for interval_index in interval_ids:
        az_time, rg_time = _inverse_geocoding_monostatic_core(gso, point, frequency_doppler_centroid, wavelength,
                                                              interval_index)
        azimuth_times.append(az_time)
        range_times.append(rg_time)
    return azimuth_times, range_times


def inverse_geocoding_bistatic(gso_rx, point: np.ndarray, frequency_doppler_centroid: float,
                               wavelength: float, gso_tx) -> (list, list):
    """Perform bistatic inverse geocoding

    :param gso_rx: general sar orbit (rx)
    :param point: a point (xyz) coordinate
    :param frequency_doppler_centroid: doppler centroid
    :param wavelength:  wavelength
    :param gso_tx: general sar orbit (tx)

    More than one solution is possible: that's why lists are returned.

    :return: azimuth_times, range_times. Two lists with the solutions.
    """

    azimuth_axis_rx = gso_rx.time_axis_array
    azimuth_axis_tx = gso_tx.time_axis_array

    try:
        dt_rx = gso_rx.dt
        dt_tx = gso_tx.dt
        dt = min(dt_tx, dt_rx)
    except RuntimeError as exc:
        raise RuntimeError("Cannot perform bistatic inverse geocoding for orbit with non-regular axis") from exc

    if dt_rx <= 0 or dt_tx <= 0:
        raise RuntimeError("Cannot perform bistatic inverse geocoding for orbit with decreasing axis")

    t_min, t_max = max(azimuth_axis_rx[0], azimuth_axis_tx[0]), min(azimuth_axis_rx[-1], azimuth_axis_tx[-1])

    if t_max <= t_min:
        raise RuntimeError('Cannot perform bistatic inverse geocoding: orbits not overlapped')

    t_ref = t_min + (t_max - t_min) / 2
    t_min = t_min - t_ref
    t_max = t_max - t_ref

    t_rx_cur = 0.
    t_tx_cur = 0.
    t_rx_old = t_max
    t_tx_old = t_max

    # t_tx_cur initialization
    position_rx = gso_rx.get_position(t_ref + t_rx_cur).squeeze()
    position_tx = gso_tx.get_position(t_ref + t_tx_cur).squeeze()
    delay_estimate = (np.linalg.norm(position_tx - point) + np.linalg.norm(position_rx - point)) / cst.LIGHT_SPEED
    t_tx_cur = t_rx_cur - delay_estimate

    interval_index_rx = -1
    interval_index_tx = -1

    iteration_counter = 0
    increment_threshold = 1.e-8
    max_iterations = 8
    relative_threshold = 2.
    increment_old = 1e+20
    increment = 1e+19

    def condition(ctr: int, incr: float, incr_old: float):
        return (ctr < max_iterations) and (incr > increment_threshold) and (incr_old / incr > relative_threshold)

    while condition(iteration_counter, increment, increment_old):
        if (t_rx_cur < t_min) or (t_rx_cur > t_max) or (t_tx_cur < t_min) or (t_tx_cur > t_max):
            raise RuntimeError('Cannot perform inverse geocoding: iterations outside the time axis')

        if (abs(t_rx_old - t_rx_cur) >= dt / 2.) or (interval_index_rx < 0):
            interval_index_rx = gso_rx.interpolator.get_interval_index(t_ref + t_rx_cur)

        position_rx = gso_rx.get_position(t_ref + t_rx_cur, interval_indexes=interval_index_rx).squeeze()
        velocity_rx = gso_rx.get_velocity(t_ref + t_rx_cur, interval_indexes=interval_index_rx).squeeze()
        acceleration_rx = gso_rx.get_acceleration(t_ref + t_rx_cur, interval_indexes=interval_index_rx).squeeze()

        if (abs(t_tx_old - t_tx_cur) >= dt / 2.) or (interval_index_tx < 0):
            interval_index_tx = gso_tx.interpolator.get_interval_index(t_ref + t_tx_cur)

        position_tx = gso_tx.get_position(t_ref + t_tx_cur, interval_indexes=interval_index_tx).squeeze()
        velocity_tx = gso_tx.get_velocity(t_ref + t_tx_cur, interval_indexes=interval_index_tx).squeeze()
        acceleration_tx = gso_tx.get_acceleration(t_ref + t_tx_cur, interval_indexes=interval_index_tx).squeeze()

        norm_vel_rx_square = np.dot(velocity_rx, velocity_rx.T)
        norm_vel_tx_square = np.dot(velocity_tx, velocity_tx.T)

        slant_range_rx = np.linalg.norm(position_rx - point)
        slant_range_tx = np.linalg.norm(position_tx - point)
        slant_range = (t_rx_cur - t_tx_cur) * cst.LIGHT_SPEED

        range_dot_vel_rx = np.dot(point - position_rx, velocity_rx)
        range_dot_vel_tx = np.dot(point - position_tx, velocity_tx)

        distance_equation_residual = slant_range_rx + slant_range_tx - slant_range
        doppler_equation_residual = -range_dot_vel_tx * slant_range_rx - range_dot_vel_rx * slant_range_tx
        doppler_equation_residual += wavelength * frequency_doppler_centroid * slant_range_rx * slant_range_tx

        df1_dt_tx = range_dot_vel_tx / slant_range_tx + cst.LIGHT_SPEED
        df1_dt_rx = range_dot_vel_rx / slant_range_rx - cst.LIGHT_SPEED
        df2_dt_tx = slant_range_rx * (norm_vel_tx_square - np.dot(point - position_tx, acceleration_tx))
        df2_dt_tx += range_dot_vel_tx / slant_range_tx * (
            range_dot_vel_rx + wavelength * frequency_doppler_centroid * slant_range_rx)
        df2_dt_rx = slant_range_tx * (norm_vel_rx_square - np.dot(point - position_rx, acceleration_rx))
        df2_dt_rx += range_dot_vel_rx / slant_range_rx * (
            range_dot_vel_tx + wavelength * frequency_doppler_centroid * slant_range_tx)

        jac_det = df1_dt_tx * df2_dt_rx - df1_dt_rx * df2_dt_tx
        inv_jac = 1. / jac_det * np.asarray([[df2_dt_rx, -df1_dt_rx], [-df2_dt_tx, df1_dt_tx]])
        delta = inv_jac @ np.asarray([distance_equation_residual, doppler_equation_residual])

        t_tx_old = t_tx_cur
        t_rx_old = t_rx_cur

        t_tx_cur = t_tx_cur - delta[0]
        t_rx_cur = t_rx_cur - delta[1]

        increment_old = increment
        increment = np.dot(delta, delta)

        iteration_counter += 1

    # lists for compatibility with monostatic version.
    t_azimuth = [t_ref + t_rx_cur]
    t_range = [t_rx_cur - t_tx_cur]
    return t_azimuth, t_range


def _inverse_geocoding_monostatic_init(gso, point: np.ndarray, frequency_doppler_centroid: float, wavelength: float):
    """Function to compute initialization of inverse geocoding

    The azimuth time we are looking for is such that the doppler equation is equal to zero.
    In this function we look for the time interval where such equation crosses zero. We compute the corresponding
    polynomial index that will be used subsequently for the precise computation of the aximuth time.
    This function evaluates the doppler equation on the azimuth time axis of the gso. It checks where it changes sign
    (from plus to minus). For each of those time positions the index of polynomial to use for interpolating in that
    neighbourhood is computed and returned. If no crossing points are found, either the first or the last polynomials
    ids are returned depending on the time instants that has smaller values of the doppler equation.

    :param gso: general sar orbit
    :param point: xyz point (ecef)
    :param frequency_doppler_centroid: a float with the doppler centroid
    :param wavelength: the wavelenth

    :return: (a list of) index of the intervals to be used for the inverse geocoding.
    """
    doppler_centroid_equation = gso.evaluate_doppler_equation(point, frequency_doppler_centroid, wavelength)

    zero_crossing_indexes = list()
    for k in range(1, doppler_centroid_equation.size):
        if doppler_centroid_equation[k] * doppler_centroid_equation[k - 1] <= 0 and doppler_centroid_equation[k] < \
                doppler_centroid_equation[k - 1]:
            zero_crossing_indexes.append(k)

    interval_index = list()
    if zero_crossing_indexes:
        azimuth_time = gso.get_interpolated_time_axis(np.asarray(zero_crossing_indexes) - 0.5)
        interval_index = gso.interpolator.get_interval_index(azimuth_time).tolist()
    else:
        if abs(doppler_centroid_equation[0]) < abs(doppler_centroid_equation[-1]):
            interval_index.append(0)
        else:
            interval_index.append(gso.interpolator.num_polynomials - 1)
    return interval_index


def _inverse_geocoding_monostatic_core(gso, point: np.ndarray, frequency_doppler_centroid: float, wavelength: float,
                                       interval_index: int):
    """Core algorithm for precise inverse geocoding

    This function finds one zero of the doppler equation as a function of the azimuth time.
    The search of the zero is restricted to the interval associated to the polynomial indexed by polynomial_index.
    Newton iterations are performed to find the exact azimuth time, once that the sensor position is known the
    computation of the range time is trivial.

    :param gso: general sar orbit
    :param point: xyz point (ecef)
    :param frequency_doppler_centroid: a float with the doppler centroid
    :param wavelength: the wavelenth
    :param interval_index: interval where to search for the solution.

    :return: (azimuth_time, range_time)

    See also _inverse_geocoding_monostatic_init
    """
    counter = 0

    polynomial_index = gso.interpolator.get_polynomial_index(interval_indexes=interval_index)[0]
    t_min = gso.time_axis_array[polynomial_index]
    t_max = gso.time_axis_array[polynomial_index + gso.interpolator.get_polynomial_order()]

    time_old = t_min
    current_azimuth_time = t_min + (t_max - t_min) / 2
    slant_range = 0
    while counter < _MAX_ITERATIONS and abs(current_azimuth_time - time_old) > _NEWTON_TOLERANCE:
        if current_azimuth_time < t_min or current_azimuth_time > t_max:
            raise RuntimeError('Cannot perform inverse geocoding')
        time_old = copy.deepcopy(current_azimuth_time)
        sat_position = gso.get_position(current_azimuth_time, interval_indexes=interval_index)[:, 0]
        sat_velocity = gso.get_velocity(current_azimuth_time, interval_indexes=interval_index)[:, 0]
        sat_acceleration = gso.get_acceleration(current_azimuth_time, interval_indexes=interval_index)[:, 0]

        point2sat = point - sat_position
        slant_range = np.linalg.norm(point2sat)
        f = np.dot(point2sat,
                   _VELOCITY_SCENE - sat_velocity) + wavelength * frequency_doppler_centroid / 2. * slant_range
        df = - np.dot(sat_velocity, _VELOCITY_SCENE - sat_velocity) - np.dot(sat_acceleration, point2sat) \
             - wavelength * frequency_doppler_centroid / 2 * np.dot(sat_velocity, point2sat) / slant_range
        current_azimuth_time -= f / df

        counter = counter + 1

    if counter == 8:
        raise RuntimeError("Exceeded max number of iterations during inverse geocoding")

    return current_azimuth_time, slant_range * 2 / cst.LIGHT_SPEED


def _check_inverse_geocoding_input(point, frequency_doppler_centroid, wavelength):
    """Checks that the input parameters for the inverse geocoding have proper types and sizes

    :param point: should be any 3-sized np.ndarray
    :param frequency_doppler_centroid: should be a float
    :param wavelength: should be a float
    """

    # checking types
    _utils.check_type(point, np.ndarray, "point")
    _utils.check_type(frequency_doppler_centroid, float, "frequency_doppler_centroid")
    _utils.check_type(wavelength, float, "wavelength")

    # size check
    if point.size != 3:
        RuntimeError("Point should be a np.ndarray of size 3")



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import numpy as np
import logging
import shutil
from scipy.signal import convolve2d
from osgeo import gdal
from equi7grid.equi7grid import Equi7Grid
from equi7grid.image2equi7grid import image2equi7grid


# biomassL2 processor imports
from biopal.fh.processing_FH import heigths_masking_and_merging
from biopal.data_operations.data_operations import (
    read_and_oversample_data,
    read_and_oversample_aux_data,
    apply_dem_flattening,
    mosaiking,
)
from biopal.utility.utility_functions import (
    Task,
    set_gdal_paths,
    choose_equi7_sampling,
    check_if_path_exists,
    check_if_geometry_auxiliaries_are_present,
    resolution_heading_correction,
    decode_unique_acquisition_id_string,
    evaluate_estimation_quality_matrix,
    save_breakpoints,
    collect_stacks_to_be_merged,
)
from biopal.geocoding.geocoding import (
    geocoding,
    geocoding_init,
)
from biopal.io.xml_io import (
    parse_input_file,
    parse_configuration_file,
    write_input_file,
    core_processing_tomo_fh,
)
from biopal.io.data_io import tiff_formatter
from biopal.screen_calibration.screen_calibration import apply_calibration_screens
from biopal.geometry.utility_geometry import compute_and_oversample_geometry_auxiliaries


from biopal.tomo.processing_TOMO import BiomassForestHeightSKPD


class TomoForestHeight(Task):
    """TOMO FH main APP TomoForestHeight

    run this APP to execute the complete Tomographic Forest Height estimation processing chain.
    
    TomoForestHeight is composed by two sub APPS automatically called in sequence  when standard launch is performed:
    StackBasedProcessingTOMOFH -> CoreProcessingTOMOFH
   
    Refer to dataset_query, StackBasedProcessingTOMOFH and CoreProcessingTOMOFH documentation for step by step run.

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the TomoForestHeight processing
    name : str
        name of the APP
        
    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query: it's the APP to be called before this APP
    StackBasedProcessingTOMOFH : it's the first of the two sub-APPs called by TomoForestHeight preocessor
    CoreProcessingTOMOFH : it's  the second of the two sub-APPs called by TomoForestHeight processor

    Examples
    --------
    Manual TOMO FH chain execution
    
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.tomo_fh.main_TOMO_FH import TomoForestHeight
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )
    >>> chain_obj = TomoForestHeight( configuration_file )
    >>> chain_obj.run( input_file_up )

    - input_file: path of the BioPAL input file
    - input_file_up: same of input_file with also the "stack_based_processing" section
    - configuration_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):

        super().__init__(configuration_file)

    def _run(self, input_file):

        # Main APP #1: Stack Based Processing
        stack_based_processing_obj = StackBasedProcessingTOMOFH(self.configuration_file)

        # Run Main APP #1: Stack Based Processing
        input_file_updated = stack_based_processing_obj.run(input_file)

        # Main APP #2: Core Processing
        tomo_fh_processing_obj = CoreProcessingTOMOFH(self.configuration_file)

        # Run Main APP #2: AGB Core Processing
        tomo_fh_processing_obj.run(input_file_updated)


class StackBasedProcessingTOMOFH(Task):
    """StackBasedProcessingTOMOFH APP 
    
    StackBasedProcessingTOMOFH APP is the first of the two sub-APPs called by TomoForestHeight processor.
    
    It performs the stack-based tomographic forest height estimation.

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file
        
    Methods
    -------
    run( input_file_path )
        run the StackBasedProcessingFH APP
    name : str
        name of the APP
        
    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query : it's the first APP to be called in the manual sequence
    CoreProcessingTOMOFH : it's the core APP that follows this APP in the call sequence

    Examples
    --------
    Manual TOMO FH chain execution
        
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.tomo_fh.main_TOMO_FH import StackBasedProcessingTOMOFH
    >>> dq_obj = dataset_query()
    >>> input_file_up1 = dq_obj.run( input_file )
    >>> sbp_obj = StackBasedProcessingTOMOFH( config_file )
    >>> input_file_up2 = sbp_obj.run( input_file_up1 )
    >>> fhcp_obj = CoreProcessingTOMOFH( config_file )
    >>> fhcp_obj.run( input_file_up2 )

    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_tomo_fh" section
    - config_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):
        super().__init__(configuration_file)

    def _run(self, input_file):

        ########################## INITIAL STEPS ##############################

        logging.info("TOMO FH: Reading chains configuration file")
        check_if_path_exists(self.configuration_file, "FILE")
        conf_params_obj = parse_configuration_file(self.configuration_file)
        input_params_obj = parse_input_file(input_file)

        ### managing output folders:
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        if conf_params_obj.processing_flags.save_breakpoints:
            breakpoints_output_folder = os.path.join(products_folder, "breakpoints")
            logging.info("TOMO FH: Breakpoints will be saved into: " + breakpoints_output_folder)
            os.makedirs(breakpoints_output_folder)

        temp_output_folder = os.path.join(products_folder, "temp")
        logging.info("TOMO FH: Temporary data folder:" + temp_output_folder + "\n")
        os.makedirs(temp_output_folder)

        equi7_sampling = choose_equi7_sampling(
            conf_params_obj.estimate_tomo_fh.product_resolution,
            input_params_obj.output_specification.geographic_grid_sampling,
        )
        e7g = Equi7Grid(equi7_sampling)
        logging.info("    EQUI7 Grid sampling used: {}".format(equi7_sampling))

        # get needed parameters from input and configuration files
        gdal_path, _ = set_gdal_paths(conf_params_obj.gdal.gdal_path, conf_params_obj.gdal.gdal_environment_path)

        ########################## INITIAL STEPS END #############################

        data_equi7_fnames = {}
        mask_equi7_fnames = {}
        ########################## STACK BASED STEPS ##############################
        for unique_stack_id, acquisitions_pf_names in input_params_obj.stack_based_processing.stack_composition.items():

            # make temporary sub-directories
            temp_output_folder_gr = os.path.join(temp_output_folder, "geocoded", unique_stack_id)
            temp_output_folder_e7 = os.path.join(temp_output_folder, "equi7", unique_stack_id)
            os.makedirs(temp_output_folder_gr)
            os.makedirs(temp_output_folder_e7)

            ### load data ( and oversample if requested and if needed )
            try:
                logging.info("TOMO FH: Data loading for stack " + unique_stack_id + "; this may take a while:")

                (beta0_calibrated, master_id, raster_info, raster_info_orig,) = read_and_oversample_data(
                    input_params_obj.dataset_query.L1C_repository,
                    acquisitions_pf_names,
                    conf_params_obj.processing_flags.enable_resampling,
                )
            except Exception as e:
                logging.error("TOMO FH: error during input data reading: " + str(e), exc_info=True)
                raise

            ### load or compute auxiliary data
            try:

                read_ref_h = (
                    not conf_params_obj.processing_flags.apply_calibration_screen
                    and conf_params_obj.processing_flags.DEM_flattening
                )
                read_cal_screens = conf_params_obj.processing_flags.apply_calibration_screen
                geometry_aux_are_present = check_if_geometry_auxiliaries_are_present(
                    input_params_obj.stack_based_processing,
                    unique_stack_id,
                    acquisitions_pf_names,
                    read_ref_h=read_ref_h,
                    read_dist=False,
                )

                if conf_params_obj.processing_flags.compute_geometry or not geometry_aux_are_present:

                    # messages for the log:
                    if conf_params_obj.processing_flags.compute_geometry:
                        logging.info("TOMO FH: calling geometry library for stack " + unique_stack_id + "\n")
                        if geometry_aux_are_present:
                            logging.warning("    geometry auxiliaries will be overwritten for stack " + unique_stack_id)
                        else:
                            logging.info("\n")
                    else:
                        logging.warning(
                            'TOMO FH: calling geometry library since AuxiliaryProductsFolder "Geometry" is empty or not complete \n'
                        )

                    (
                        _,
                        _,
                        ellipsoid_slope,
                        _,
                        _,
                        _,
                        sar_geometry_master,
                    ) = compute_and_oversample_geometry_auxiliaries(
                        input_params_obj.dataset_query.L1C_repository,
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        master_id,
                        conf_params_obj.processing_flags.enable_resampling,
                        comp_ref_h=read_ref_h,
                        comp_dist=False,
                        force_ellipsoid=True,
                    )

                    (
                        ecef_grid,
                        off_nadir_angle_rad,
                        slope,
                        kz,
                        reference_height,
                        _,
                        _,
                    ) = compute_and_oversample_geometry_auxiliaries(
                        input_params_obj.dataset_query.L1C_repository,
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        master_id,
                        conf_params_obj.processing_flags.enable_resampling,
                        comp_ref_h=read_ref_h,
                        comp_dist=False,
                        sar_geometry_master=sar_geometry_master,
                    )

                    logging.info("Geometry library: correcting geometry global / local reference")
                    slope = slope - ellipsoid_slope
                    for swath_id in kz.keys():
                        kz[swath_id] = (
                            kz[swath_id]
                            * np.sin(off_nadir_angle_rad[master_id])
                            / np.sin(off_nadir_angle_rad[master_id] - ellipsoid_slope)
                        )
                    off_nadir_angle_rad[master_id] = off_nadir_angle_rad[master_id] - ellipsoid_slope
                    del ellipsoid_slope

                    logging.info("TOMO FH: ...geometry auxiliaries computation done.")

                else:

                    logging.info(
                        "TOMO FH: geometry auxiliaries are provided from user, so they are now being loaded and not computed, for stack "
                        + unique_stack_id
                        + "\n"
                    )

                    (
                        ecef_grid,
                        off_nadir_angle_rad,
                        slope,
                        kz,
                        reference_height,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                    ) = read_and_oversample_aux_data(
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        conf_params_obj.processing_flags.enable_resampling,
                        raster_info_orig,
                        read_ref_h=read_ref_h,
                        read_dist=False,
                    )
                    logging.info("TOMO FH: ...geometry auxiliaries loading done.")

                # read the rest of auxiliaries which are notpart of the geometry library:
                if read_cal_screens:

                    logging.warning("TOMO FH: loading calibration screens \n")

                    (
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        cal_screens,
                        cal_screens_raster_info,
                        _,
                        _,
                        _,
                    ) = read_and_oversample_aux_data(
                        input_params_obj.stack_based_processing,
                        unique_stack_id,
                        acquisitions_pf_names,
                        conf_params_obj.processing_flags.enable_resampling,
                        raster_info_orig,
                        read_cal_screens=read_cal_screens,
                        read_ecef=False,
                        read_off_nadir=False,
                        read_slope=False,
                        read_kz=False,
                        read_ref_h=False,
                        read_dist=False,
                    )
                    logging.info("...done")

            except Exception as e:
                logging.error(
                    "TOMO FH: error during auxiliary data computation and/or loading: " + str(e), exc_info=True,
                )
                raise

            ### Screen calibration (ground steering)
            try:
                if conf_params_obj.processing_flags.apply_calibration_screen:
                    logging.info("TOMO FH: applying calibration screen...")
                    beta0_calibrated = apply_calibration_screens(
                        beta0_calibrated, raster_info, cal_screens, cal_screens_raster_info, master_id,
                    )
                    logging.info("...done.\n")

                elif conf_params_obj.processing_flags.DEM_flattening:
                    logging.info("TOM FH: DEM flattening... ")
                    beta0_calibrated = apply_dem_flattening(
                        beta0_calibrated, kz, reference_height, master_id, raster_info
                    )
                    logging.info("...done.\n")

            except Exception as e:
                logging.error(
                    "TOMO FH: error during screen calibration or DEM flattening." + str(e), exc_info=True,
                )
                raise

            ### compute mean incidence angle
            look_angle_rad = np.nanmean(off_nadir_angle_rad[master_id])  # 0.4886921905584123
            logging.info("TOMO FH: incidence angle used is {} [deg] \n".format(np.rad2deg(look_angle_rad)))

            ### mean of off nadir and slope over final resolution
            windtm_x = np.int(
                np.round(conf_params_obj.estimate_tomo_fh.product_resolution / raster_info.pixel_spacing_az / 2) * 2 + 1
            )
            windtm_y = np.int(
                np.round(
                    conf_params_obj.estimate_tomo_fh.product_resolution
                    / (raster_info.pixel_spacing_slant_rg / np.sin(look_angle_rad))
                    / 2
                )
                * 2
                + 1
            )

            logging.info("TOMO FH: computing mean of off nadir based on final product resolution...")

            # only mster is used in the code, do not convolve all the stacs, it is useless
            off_nadir_angle_rad[master_id] = convolve2d(
                off_nadir_angle_rad[master_id], np.ones((windtm_y, windtm_x)) / windtm_y / windtm_x, mode="same",
            )
            logging.info("...done.")

            logging.info("TOMO FH: computing mean of slope based on final product resolution...")

            slope = convolve2d(slope, np.ones((windtm_y, windtm_x)) / windtm_y / windtm_x, mode="same")
            logging.info("...done.")

            # covariance estimation window size, it may be modified by an internal flag in case of air-plane geometry
            cov_est_window_size = conf_params_obj.estimate_tomo_fh.product_resolution

            if conf_params_obj.processing_flags.multilook_heading_correction:
                _, heading_deg, _, _, _, _ = decode_unique_acquisition_id_string(unique_stack_id + "_BSL_00")

                cov_est_window_size = resolution_heading_correction(cov_est_window_size, heading_deg)

            logging.info("Covariance Estimation Window Size used: {} [m]".format(cov_est_window_size))

            ### height estimation
            logging.info("TOMO FH: " + unique_stack_id + ": performing heigth estimation...")
            try:

                vertical_vector = np.arange(
                    conf_params_obj.estimate_tomo_fh.vertical_range.minimum_height,
                    conf_params_obj.estimate_tomo_fh.vertical_range.maximum_height
                    + conf_params_obj.estimate_tomo_fh.vertical_range.sampling,
                    conf_params_obj.estimate_tomo_fh.vertical_range.sampling,
                )

                (estimated_height, power_peak, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = BiomassForestHeightSKPD(
                    beta0_calibrated,
                    cov_est_window_size,
                    raster_info.pixel_spacing_slant_rg,
                    raster_info.pixel_spacing_az,
                    look_angle_rad,
                    raster_info.carrier_frequency_hz,
                    raster_info.range_bandwidth_hz,
                    kz,
                    vertical_vector,
                    conf_params_obj.estimate_tomo_fh,
                )

                estimated_height = estimated_height * (
                    1
                    - np.tan(slope[rg_vec_subs, :][:, az_vec_subs])
                    / np.tan(off_nadir_angle_rad[master_id][rg_vec_subs, :][:, az_vec_subs])
                )

            except Exception as e:
                logging.error("FH: error during height estimation: " + str(e), exc_info=True)
                raise
            del beta0_calibrated

            ### Placemark for the quality estimation to be defined
            try:
                quality_layer_sr = evaluate_estimation_quality_matrix(estimated_height.shape)

            except Exception as e:
                logging.error("FH: error during estimation quality layer evaluation: " + str(e), exc_info=True)
                raise

            ### Interpolate it over a regular lat lon grid (with grid data): generate a regular grid for the interpolation by using Max and Min lat lon from the ECEFGRID_LLH (make the vectors a bit longer), for the grid steps use the minimum steps from the ECEFGRID_LLH)
            logging.info(unique_stack_id + ": Geocoding data...")
            try:

                # initialize the geocoding
                min_spacing_m = min(
                    subs_F_a * raster_info.pixel_spacing_az, subs_F_r * raster_info.pixel_spacing_slant_rg,
                )
                min_spacing_m = min(min_spacing_m, equi7_sampling)

                logging.info("Geocoding spacing set to {} [m]".format(min_spacing_m))

                (
                    lon_in,
                    lat_in,
                    alt_in,
                    lonMeshed_out,
                    latMeshed_out,
                    valid_values_mask,
                    lon_regular_vector,
                    used_lon_step,
                    lat_regular_vector,
                    used_lat_step,
                ) = geocoding_init(ecef_grid, rg_vec_subs, az_vec_subs, min_spacing_m)

                # geocode the FH
                logging.info(unique_stack_id + ": geocoding the estimated height...")
                data_ground = geocoding(
                    estimated_height, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask,
                )
                logging.info("...done.\n")

                # geocode the FH quality layer
                logging.info(unique_stack_id + ": geocoding the estimated height quality layer...")
                quality_layer_ground = geocoding(
                    quality_layer_sr, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask,
                )
                logging.info("...done.\n")

            except Exception as e:
                logging.error("TOMO FH: error during geocoding: " + str(e), exc_info=True)
                raise

            ### saving breakpoints
            if conf_params_obj.processing_flags.save_breakpoints:
                logging.info("TOMO FH: saving mail results (in slant range geometry) on " + breakpoints_output_folder)
                post_string = "_SR_" + unique_stack_id

                breakpoint_names = ["estimated_height" + post_string]

                save_breakpoints(breakpoints_output_folder, breakpoint_names, [estimated_height])
                logging.info("...done.\n")
            del estimated_height

            ### creating mask to exclude estimation not valid values:
            condition_curr = np.logical_and(
                data_ground > conf_params_obj.estimate_tomo_fh.estimation_valid_values_limits[0],
                data_ground < conf_params_obj.estimate_tomo_fh.estimation_valid_values_limits[1],
            )
            estimation_mask_ground = np.where(condition_curr, True, False)
            estimation_mask_ground[np.isnan(data_ground)] = False
            # also casted to float, for incapsulation in a geotiff
            estimation_mask_ground.astype(float)

            ### create GEOTIFF of all the layers (estimation, mask ):
            logging.info(unique_stack_id + ": formatting data to GEOTIFF...")
            try:

                data_ground_fname = os.path.join(temp_output_folder_gr, "FH.tif")
                valid_values_mask_ground_fname = os.path.join(temp_output_folder_gr, "fnf.tif")

                upper_left_easting_coord = lon_regular_vector[0]  # i.e. horizontal
                sampling_step_east_west = used_lon_step
                upper_left_northing_coord = lat_regular_vector[0]  # i.e. vertical
                sampling_step_north_south = used_lat_step
                geotransform = [
                    upper_left_easting_coord,
                    sampling_step_east_west,
                    0,
                    upper_left_northing_coord,
                    0,
                    sampling_step_north_south,
                ]

                # forest height geotiff formatting
                tiff_formatter(
                    [data_ground, quality_layer_ground],
                    data_ground_fname,
                    geotransform,
                    gdal_data_format=gdal.GDT_Float32,
                    multi_layers_tiff=True,
                )

                # forest height temporary mask geotiff formatting (this mask should be merged with the fnf mask, generated after )
                tiff_formatter(
                    estimation_mask_ground,
                    valid_values_mask_ground_fname,
                    geotransform,
                    gdal_data_format=gdal.GDT_Float32,
                )

                logging.info("...done.\n")

            except Exception as e:
                logging.error("TOMO FH: error during GEOTIFF formatting: " + str(e), exc_info=True)
                raise

            ### formatting data to EQUI7
            logging.info(unique_stack_id + ": formatting into EQUI7 grid...")
            try:

                equi7_data_outdir = os.path.join(temp_output_folder_e7, "data")
                equi7_mask_outdir = os.path.join(temp_output_folder_e7, "fnf")

                # in general from here the Equi7 can output multiple tiles, which file names are stored in the output list ( wrapped here in a dict for the stack )
                data_equi7_fnames[unique_stack_id] = image2equi7grid(
                    e7g,
                    data_ground_fname,
                    equi7_data_outdir,
                    gdal_path=gdal_path,
                    inband=None,
                    subgrid_ids=None,
                    accurate_boundary=False,
                    withtilenamesuffix=False,
                    resampling_type="bilinear",
                    tile_nodata=np.nan,
                )
                mask_equi7_fnames[unique_stack_id] = image2equi7grid(
                    e7g,
                    valid_values_mask_ground_fname,
                    equi7_mask_outdir,
                    gdal_path=gdal_path,
                    inband=None,
                    subgrid_ids=None,
                    accurate_boundary=False,
                    withtilenamesuffix=False,
                    resampling_type="bilinear",
                    tile_nodata=np.float(0),
                )

            except Exception as e:
                logging.error("TOMO FH: error during EQUI7 formatting: " + str(e), exc_info=True)
                raise

            if not os.path.exists(data_equi7_fnames[unique_stack_id][0]):
                error_message = "EQUI7 grid has not been generated, output is absent "
                logging.error(error_message)
                raise

            logging.info("...done.\n")

        # write the input file with the sections needed by the Core Processing FH APP:
        out_input_file_xml = os.path.join(
            os.path.dirname(input_params_obj.output_specification.output_folder), "Input_File_CoreProcessingFH.xml"
        )
        input_params_obj.core_processing_tomo_fh = fill_core_processing_tomo_fh_obj(
            input_params_obj, data_equi7_fnames, mask_equi7_fnames,
        )
        write_input_file(input_params_obj, out_input_file_xml)

        ######################## STACK BASED STEPS END. ###########################
        return out_input_file_xml


class CoreProcessingTOMOFH(Task):
    """CoreProcessingTOMOFH APP 
    
    CoreProcessingTOMOFH APP is the second of the two sub-APPs called by TomoForestHeight processor.

    It performs merging ascending with descending processed stacks and
    mosaiking together differtent processed tiles of the output grid

    Attributes
    ----------
    configuration_file : str
        path of the Configuration_File.xml file

    Methods
    -------
    run( input_file_path )
        run the CoreProcessingTOMOFH APP
    name : str
        name of the APP
        
    See Also
    --------
    biopal.dataset_query.dataset_query.dataset_query : it's the first APP to be called in the manual sequence
    StackBasedProcessingTOMOFH : it’s the APP which prepares the input for this APP

    Examples
    -------- 
    Manual TOMO FH chain execution
        
    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> from biopal.tomo_fh.main_TOMO_FH import StackBasedProcessingTOMOFH
    >>> dq_obj = dataset_query()
    >>> input_file_up1 = dq_obj.run( input_file )
    >>> sbp_obj = StackBasedProcessingTOMOFH( config_file )
    >>> input_file_up2 = sbp_obj.run( input_file_up1 )
    >>> fhcp_obj = CoreProcessingTOMOFH( config_file )
    >>> fhcp_obj.run( input_file_up2 )

    - input_file: path of the BioPAL input file
    - input_file_up1: same of input_file with also the "stack_based_processing" section   
    - input_file_up2: same of input_file_up1 with also the "core_processing_tomo_fh" section
    - config_file: path of the BioPAL configuration file
    """

    def __init__(self, configuration_file):

        super().__init__(configuration_file)

    def _run(self, input_file):

        # FH: Reading chains configuration files
        logging.info("TOMO FH: Reading chains configuration files")
        check_if_path_exists(self.configuration_file, "FILE")
        conf_params_obj = parse_configuration_file(self.configuration_file)
        input_params_obj = parse_input_file(input_file)

        # managing folders
        products_folder = os.path.join(input_params_obj.output_specification.output_folder, "Products")
        temp_output_folder = os.path.join(products_folder, "temp")

        ######################## NOT STACK BASED STEPS ############################

        stacks_to_merge_dict = collect_stacks_to_be_merged(input_params_obj.stack_based_processing.stack_composition)

        try:

            logging.info("TOMO FH: merging ascending with descending stacks....\n")

            merged_data_fnames, merging_folder = heigths_masking_and_merging(
                input_params_obj.core_processing_tomo_fh.data_equi7_fnames,
                input_params_obj.core_processing_tomo_fh.mask_equi7_fnames,
                stacks_to_merge_dict,
            )

            logging.info("...done.\n")

        except Exception as e:
            logging.error(e, exc_info=True)
            raise

        try:

            logging.info("TOMO FH: mosaiking equi7 tiles....\n")

            output_folder = os.path.join(products_folder, "global_FH")

            mosaiking(merging_folder, output_folder)

            logging.info("...done.\n")

        except Exception as e:
            logging.error(e, exc_info=True)
            raise

        if conf_params_obj.processing_flags.delete_temporary_files:
            shutil.rmtree(temp_output_folder)

        logging.info("TOMO FH: Forest Height estimation ended correctly.\n")


def fill_core_processing_tomo_fh_obj(input_params_obj, data_equi7_fnames, mask_equi7_fnames):

    """
    Internal function called by the StackBasedProcessingTOMOFH APP:
        
        the StackBasedProcessingTOMOFH APP fills the structure
        to be written into the xml input file for the next APP, which is the 
        core processing for the TOMO FH.
        The returned object "core_processing_TOMOFH_obj" contains the paths of 
        computed data and masks in equi7 format
        
        Usage of the returned object:
            The returned object can be added to the input_params_obj and it 
            can be written to disk if needed, as (core_processing_tomofh_obj is 
                                                  overwritten by this command, if already present):
                - input_params_obj.core_processing_tomo_fh_obj = core_processing_tomo_fh_obj
                - write_input_file(input_params_obj, input_file_xml)
    """

    core_processing_tomofh_obj = core_processing_tomo_fh(data_equi7_fnames, mask_equi7_fnames,)

    return core_processing_tomofh_obj



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Constants module
----------------

Example of usage:

.. code-block:: python

    import arepytools.constants as cst
    print(cst.LIGHT_SPEED)
"""

# Speed of light
LIGHT_SPEED = 299792458.0
"""
Speed of light in vacuum (m/s).
"""

# Units of measure
SECOND_STR = 's'
"""
Second symbol.
"""

HERTZ_STR = 'Hz'
"""
Hertz symbol.
"""

JOULE_STR = 'j'
"""
Joule symbol.
"""

DEGREE_STR = 'deg'
"""
Degree symbol.
"""

RAD_STR = 'rad'
"""
Radian symbol.
"""

UTC_STR = 'Utc'
"""
Coordinated Universal Time abbreviation.
"""

METER_STR = 'm'
"""
Meter symbol.
"""

# Metric prefixes
KILO = 1e3
"""
Number of units in one kilounit (kilounits-to-units conversion factor).
"""

MEGA = KILO**2
"""
Number of units in one megaunit (megaunits-to-units conversion factor).
"""

GIGA = KILO**3
"""
Number of units in one gigaunit (gigaunits-to-units conversion factor).
"""

MILLI = KILO**-1
"""
Number of units in one milliunit (milliunits-to-units conversion factor).
"""

MICRO = KILO**-2
"""
Number of units in one microunit (microunits-to-units conversion factor).
"""

NANO = KILO**-3
"""
Number of units in one nanounit (nanounits-to-units conversion factor).
"""

PICO = KILO**-4
"""
Number of units in one picounit (picounits-to-units conversion factor).
"""



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
ProductFolder module
--------------------
"""
import re
import os
import lxml.etree as etree
from arepytools.io import read_metadata
from ._utils import EOpenMode
from .metadata import RasterInfo
from . import channel

_re_pattern = '_(?!0000)[0-9]{4}'
_header_extension = '.xml'


class ProductFolder:
    """
    ProductFolder class
    """

    def __init__(self, dir_path, open_mode=EOpenMode('r')):

        self.__open_mode = EOpenMode(open_mode)
        self.__pf_dir_path = os.path.abspath(dir_path)
        self.__pf_name = os.path.basename(dir_path)
        self.__channels = list()

        if EOpenMode(open_mode) == EOpenMode('r'):
            if not self.is_productfolder(dir_path):
                raise IsNotAProductFolder
            self.__pf_read(dir_path, open_mode)

        elif EOpenMode(open_mode) == EOpenMode('x') or EOpenMode(open_mode) == EOpenMode('w'):
            if os.path.exists(dir_path):
                raise RuntimeError(
                    "Cannot initialize {} on {}: path already exists".format(self.__class__.__name__, dir_path))

            # Initialize the product folder
            self.__pf_generate(dir_path)
        else:
            raise NotImplementedError("{} mode not supported".format(EOpenMode(open_mode)))

    @property
    def open_mode(self):
        return self.__open_mode

    @property
    def pf_name(self):
        return self.__pf_name

    @property
    def pf_dir_path(self):
        return self.__pf_dir_path

    def __pf_generate(self, dir_path):
        # Create the directory
        os.mkdir(dir_path)
        # Create the manifest
        self.__create_manifest(dir_path)

    @staticmethod
    def __create_manifest(dir_path):
        manifest = etree.Element("AresysProductManifest")
        manifest.set('Version', '2')
        etree.SubElement(manifest, "ProductDescription").text = 'ProductFolder initialized by ArePyTools'
        tree = etree.ElementTree(manifest)
        tree.write(os.path.join(dir_path, 'aresys_product'), pretty_print=True, xml_declaration=True, encoding="utf-8")

    def __pf_read(self, dir_path, open_mode):
        headers = self._get_header_files_in_dir(dir_path)
        raster_extension = self.get_data_file_extension(dir_path)
        for header in headers:
            metadata = read_metadata(os.path.join(dir_path, header))
            raster_filename = header[0:-len(_header_extension)] + raster_extension
            raster_path = os.path.join(dir_path, raster_filename)
            self.__channels.append(channel.Channel(raster_path, metadata, open_mode))

    def append_channel(self, lines, samples, data_type, header_offset=0, row_prefix=0, byte_order='LITTLEENDIAN'):
        """
        Append a channel to the ProductFolder

        :param lines: number of lines
        :param samples: number of samples
        :param data_type: raster data type (see :class:`arepytools.io.metadata.ECellType`)
        :param header_offset: header offset in the raster
        :param row_prefix: row prefix in the raster
        :param byte_order: byte order of the raster (see :class:`arepytools.io.metadata.EByteOrder`)
        """
        if self.open_mode == EOpenMode('r'):
            raise ReadOnlyProductFolder
        # Define the channel file name
        current_num_of_channels = self.get_number_channels()
        _, pf_name = os.path.split(self.pf_dir_path)
        suffix = "_{:0>4}".format(current_num_of_channels + 1)
        chan_file_path = os.path.join(self.pf_dir_path, pf_name + suffix)

        # Create the metadata
        _, file_name = os.path.split(chan_file_path)
        current_raster_info = RasterInfo(lines, samples, data_type, file_name, header_offset, row_prefix, byte_order)

        # Create the new channel
        current_channel = channel.Channel.from_raster_info(chan_file_path, current_raster_info, self.open_mode)

        # Add the channel to the product folder
        self.__channels.append(current_channel)

    @staticmethod
    def _get_header_files_in_dir(dir_path) -> list:
        files = os.listdir(dir_path)
        pf_name = os.path.split(dir_path)[-1]
        header_list = [f for f in files if
                       re.fullmatch("{}".format(pf_name) + _re_pattern + _header_extension, f) is not None]
        header_list.sort()
        return header_list

    @staticmethod
    def _get_raster_files_in_dir(dir_path) -> list:
        files = os.listdir(dir_path)
        pf_name = os.path.split(dir_path)[-1]
        raster_list = [f for f in files if re.fullmatch("{}".format(pf_name) + _re_pattern, f) is not None]
        raster_list.sort()
        return raster_list

    @classmethod
    def is_productfolder(cls, dir_path) -> bool:
        """
        Verify if the folder at the path in input is a ProductFolder

        :param dir_path: path of the folder to check
        :return: True if the folder is a ProductFolder False instead
        """
        if not os.path.isdir(dir_path):
            return False

        files = os.listdir(dir_path)

        # Check manifest existence
        if files.count('aresys_product') != 1:
            return False

        # Exist the raster for each header
        raster_extension = cls.get_data_file_extension(dir_path)
        for header in cls._get_header_files_in_dir(dir_path):
            raster_filename = header[0:-len(_header_extension)] + raster_extension
            if files.count(raster_filename) != 1:
                return False

        # Exist the header for each raster
        for raster_file in cls._get_raster_files_in_dir(dir_path):
            if files.count(raster_file + _header_extension) != 1:
                return False

        return True

    @classmethod
    def get_data_file_extension(cls, dir_path) -> str:
        manifest_filename = os.path.join(dir_path, 'aresys_product')
        manifest_root = etree.parse(manifest_filename).getroot()
        extension_node = manifest_root.find('DataFileExtension')

        if extension_node is not None and extension_node.text is not None:
            return extension_node.text
        return ''

    def get_number_channels(self) -> int:
        """
        Get the number of channels in the ProductFolder

        :return: number of channels
        """
        return len(self.__channels)

    def read_binary_header(self, channel_index):
        """
        Read the header of the raster of a channel of the ProductFolder

        :param channel_index: index of the channel to read
        :return: header in bytes
        """
        return self.__channels[channel_index].read_binary_header()

    def write_binary_header(self, channel_index, header):
        """
        Write data to a channel of the ProductFolder

        :param channel_index: index of the channel to write
        :param header: header in bytes
        """
        if self.open_mode == EOpenMode('r'):
            raise ReadOnlyProductFolder
        self.__channels[channel_index].write_binary_header(header)

    def read_row_prefix(self, channel_index, line_index):
        """
        Read the row prefix of a given line of a channel of the ProductFolder

        :param channel_index: index of the channel
        :param line_index: raster line index
        :return: row prefix as byte sequence
        """
        return self.__channels[channel_index].read_row_prefix(line_index)

    def write_row_prefix(self, channel_index, line_index, row_prefix):
        """
        Write the row prefix of a given line of a channel of the ProductFolder

        :param channel_index: index of the channel
        :param line_index: raster line index
        :param row_prefix: row prefix as bytes sequence
        """
        if self.open_mode == EOpenMode('r'):
            raise ReadOnlyProductFolder
        self.__channels[channel_index].write_row_prefix(line_index, row_prefix)

    def read_data(self, channel_index, block_to_read=None):
        """
        Read the raster of a channel of the ProductFolder

        :param channel_index: index of the channel to read
        :param block_to_read: portion of the raster to read [start line, start sample, number of lines, number of
            samples]
        :return: read data
        """
        data = channel.Channel.read_data(self.__channels[channel_index], block_to_read)
        return data

    def write_data(self, channel_index, data, start_point=(0, 0)):
        """
        Write data to a channel of the ProductFolder

        :param channel_index: index of the channel to write
        :param data: data to write
        :param start_point: coordinates of the first pixel to write
        """
        if self.open_mode == EOpenMode('r'):
            raise ReadOnlyProductFolder
        self.__channels[channel_index].write_data(data, start_point)

    def write_metadata(self, channel_index):
        """
        Write the metadata to the xml for a channel of the ProductFolder

        :param channel_index: index of the channel to write
        """
        if self.open_mode == EOpenMode('r'):
            raise ReadOnlyProductFolder
        self.__channels[channel_index].write_metadata()

    def get_channel(self, channel_index):
        """
        Get a channel of the ProductFolder

        :param channel_index: index of the channel to return
        :return: :class:`arepytools.io.metadata.channel.Channel` -- Channel for the index in input
        """
        return self.__channels[channel_index]


class IsNotAProductFolder(ValueError):
    """Invalid product folder error"""

    def __init__(self):
        super().__init__('The directory in input is not a valid ProductFolder')


class ReadOnlyProductFolder(IOError):
    """Read only product folder error"""

    def __init__(self):
        super().__init__('Operation not allowed: product folder open in read mode')
